#include "EMVtk.h"
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>

EMVtk::EMVtk(const char* filename)
    : fname(filename)
{
    in = 0;
    type = -1;
    nx = 0;
    ny = 0;
    nz = 0;
    originx = 0;
    originy = 0;
    originz = 0;
    spacingx = 0;
    spacingy = 0;
    spacingz = 0;
}

EMVtk::EMVtk(FILE* infile)
    : in(infile)
{
    fname = 0;
    type = -1;
    nx = 0;
    ny = 0;
    nz = 0;
    originx = 0;
    originy = 0;
    originz = 0;
    spacingx = 0;
    spacingy = 0;
    spacingz = 0;
}

EMVtk::~EMVtk()
{
    if (in) {
	fclose(in);
	in = 0;
    }
}

static int is_empty_line(const char* line)
{
    if (line) {
	const char* tmp = line;
	for (; !tmp; tmp++) {
	    if(!isspace(tmp[0])) {
		return 0;
	    }
	}
    }
    return 1;
}

	
int EMVtk::init()
{
    if(fname) {
	in = fopen(fname, "r");
	if (!in) {
	    fprintf(stderr, "Error: cannot open file '%s'\n", fname);
	    return 1;
	}
    }
    
    char buf[1024];
    int bufsz = sizeof(buf);
    const char* magic = "# vtk DataFile";
    
    if (fgets(buf, bufsz, in)) {
	if (strncmp(buf, magic, strlen(magic)) != 0) {
	    return 1;
	}
    }
    else {
	return 1;
    }

    fgets(title, sizeof(title), in);
    char vtktype[32];
    fgets(vtktype, sizeof(vtktype), in);
    
    if (strncmp(vtktype, "ASCII", strlen("ASCII")) == 0) {
	type = VTK_ASCII;
    }
    else if (strncmp(vtktype, "BINARY", strlen("BINARY")) == 0) {
	type = VTK_BINARY;
    }
    else {
	fprintf(stderr, "Error: invalid VTK type: '%s'\n", vtktype);
	return 1;
    }

    if (type == VTK_BINARY) {
	fprintf(stderr, "WARNING: can only read ASCII VTK file\n");
	return 1;
    }

    fgets(buf, bufsz, in);
    if(is_empty_line(buf)) {
	fgets(buf, bufsz, in);
    }
    
    if (fgets(buf, bufsz, in)) {
	if(sscanf(buf, "DIMENSIONS %d %d %d", &nx, &ny, &nz) != 3) {
	    return 1;
	}
    }
    else {
	return 1;
    }

    fgets(buf, bufsz, in);
    if(sscanf(buf, "ORIGIN %f %f %f", &originx, &originy, &originz) != 3) {
	return 1;
    }

    fgets(buf, bufsz, in);
    if(sscanf(buf, "SPACING %f %f %f", &spacingx, &spacingy, &spacingz) != 3) {
	return 1;
    }

    if (spacingx != spacingy || spacingx != spacingz || spacingy != spacingz) {
	fprintf(stderr, "Error: cannot handle non-uniform spacing VTK file\n");
	return 1;
    }
    
    fgets(buf, bufsz, in);
    if (is_empty_line(buf)) {
	fgets(buf, bufsz, in);
    }
       
    int nitems = 0;
    sscanf(buf, "POINT_DATA %d", &nitems);
    if (nitems != (nx*ny*nz)) {
	fprintf(stderr, "Error: number of point data is invalid\n");
	return 1;
    }
    
    fgets(buf, bufsz, in);
    buf[strlen(buf)-1] = '\0';
    if(strncmp(buf, "SCALARS density float", strlen("SCALARS density float")) != 0) {
	fprintf(stderr, "Error: invalid format '%s'\n", buf);
	return 1;
    }

    fgets(buf, bufsz, in);
    buf[strlen(buf)-1] = '\0';
    if(strncmp(buf, "LOOKUP_TABLE default", strlen("LOOKUP_TABLE default")) != 0) {
	fprintf(stderr, "Error: invalid format '%s'\n", buf);
	return 1;
    }
    fgets(buf, bufsz, in);
    
#ifdef DEBUG
    printf("vtk nx = %d, ny = %d, nz = %d\n", nx, ny, nz);
#endif
    return 0;
}


int EMVtk::read_data(float** p_data)
{
    if (!p_data || !(*p_data)) {
	fprintf(stderr, "Error: null pointer in read_data()\n");
	return 1;
    }

    int bufsz = nx*16;
    char* buf = new char[bufsz];
    
    for (int i = 0; i < nz; i++) {
	for (int j = 0; j < ny; j++) {
	    if (fgets(buf, bufsz, in)) {
		char* s1 = strtok(buf, "\t");
		int k = 0;
		(*p_data)[i*nx*ny+j*nx+k] = atoi(s1);
		k++;
		
		while((s1 = strtok(0, "\t"))) {
		    (*p_data)[i*nx*ny+j*nx+k] = atoi(s1);
		    k++;
		}

		assert((k-1) == nx);
	    }
	    else {
		return 1;
	    }
	}
	
    }
    delete [] buf;
    buf = 0;
    return 0;
}

