#include "PyEuler.h"
#include "PyPickle.h"

#ifdef DEBUG
#include "assert.h"
#endif

static PyEuler toPyEuler(Euler e)
{
    return *(static_cast<PyEuler*>(&e));
}


PyEuler::PyEuler()
    : Euler()
{

}

PyEuler::PyEuler(float a1, float a2, float a3)
    : Euler(a1, a2, a3)
{
}

PyEuler::PyEuler(float a1, float a2, float a3, int typ) :
    Euler(a1, a2, a3, typ)
{
}

PyEuler::PyEuler(float a1, float a2, float a3, float a0, int typ) :
    Euler(a1, a2, a3, a0, typ)
{
}

PyEuler::PyEuler(PyList m_list)
{
    py_setAngle(m_list);
    init();
}

PyEuler::~PyEuler()
{
}


void PyEuler::py_setAngle(PyList m_list)
{
#if DEBUG
    assert(m_list.size() == EULER_MATRIX_SIZE);
#endif

    float* m_array = new float[EULER_MATRIX_SIZE];
    list2array<float>(m_list, m_array);
    setAngle(m_array);
    array2list<float>(m_array, m_list);
    delete [] m_array;
    m_array = 0;
}

void PyEuler::py_setSym(const char *symmetry)
{
    setSym(const_cast<char*>(symmetry));
}

PyList PyEuler::py_mat()
{
    float* m_array = mat();
    PyList m_list(EULER_MATRIX_SIZE);
    
    array2list<float>(m_array, m_list);
    return m_list;
}

void PyEuler::py_print(const char *id, FILE* out)
{
    print(const_cast<char*>(id), out);
}

void PyEuler::py_printMatrix(const char *id, FILE* out)
{
    printMatrix(const_cast<char*>(id), out);
}

PyEuler PyEuler::py_inverse()
{
    return toPyEuler(inverse());
}

PyEuler PyEuler::py_EulerMult(PyEuler bc)
{
    return toPyEuler(bc);
}

PyEuler PyEuler::py_FirstSym()
{
    return toPyEuler(FirstSym());
}

PyEuler PyEuler::py_NextSym()
{
    return toPyEuler(NextSym());
}


PyEuler PyEuler::py_SymN(int n)
{
    return toPyEuler(SymN(n));
}


PyEuler operator*(PyEuler lhs, PyEuler rhs)
{
    Euler e1 = lhs;
    Euler e2 = rhs;
    Euler r =  e1 * e2;
    return toPyEuler(r);
}

PyEuler operator/(PyEuler lhs, PyEuler rhs)
{
    Euler e1 = lhs;
    Euler e2 = rhs;
    Euler r = e1 / e2;
    return toPyEuler(r);
}

bool operator!=(PyEuler lhs, PyEuler rhs)
{
    Euler e1 = lhs;
    Euler e2 = rhs;
    return (e1 != e2);
}


bool operator==(PyEuler lhs, PyEuler rhs)
{
    return !(lhs != rhs);
}


BOOST_PYTHON_BEGIN_CONVERSION_NAMESPACE
static python_extension_class_converters<PyEuler> _PyEuler_type;
BOOST_PYTHON_END_CONVERSION_NAMESPACE


enum EulerState {
    EULER_ALT,
    EULER_AZ,
    EULER_PHI,
    EULER_MAT,
    EULER_DICT,
    EULER_NUM_STATES
};

// use constructor: PyEuler(float a1, float a2, float a3)
python::ref Euler_getinitargs(PyEuler& euler)
{
    python::tuple result(3);
    result.set_item(0, euler.alt());
    result.set_item(1, euler.az());
    result.set_item(2, euler.phi());
    
    return result.reference();
}



python::ref Euler_getstate(const python::tuple& args,
			   const python::dictionary& keywords)
{
    check_args(args, keywords, 1);

    PyEuler& euler = BOOST_PYTHON_CONVERSION::from_python(args[0].get(),
							  python::type<PyEuler&>());
    
    python::tuple result(EULER_NUM_STATES);
    
    result.set_item(EULER_ALT, euler.alt());
    result.set_item(EULER_AZ, euler.az());
    result.set_item(EULER_PHI, euler.phi());

    PyList m_list(9);
    float* m_array = euler.mat();
    array2list<float>(m_array, m_list);

    result.set_item(EULER_MAT, m_list);

    python::ref mydict = python::getattr(args[0], "__dict__");
    result.set_item(EULER_DICT, mydict);
    
    return result.reference();

}

PyObject* Euler_setstate(const python::tuple& args,
			 const python::dictionary& keywords)
{
    set_dict(args, keywords, EULER_NUM_STATES);

    PyEuler& euler = BOOST_PYTHON_CONVERSION::from_python(args[0].get(),
							  python::type<PyEuler&>());    
    python::tuple state = BOOST_PYTHON_CONVERSION::from_python(args[1].get(),
							       python::type<python::tuple>());

    euler.setAngle(get_float(state[EULER_ALT]),
		   get_float(state[EULER_AZ]),
		   get_float(state[EULER_PHI]));
    
    PyList m_list = BOOST_PYTHON_CONVERSION::from_python(state[EULER_MAT].get(),
							 python::type<PyList>());
#ifdef DEBUG
    assert(m_list.size() == 9);
#endif    
    float* m_array = new float[m_list.size()];
    list2array<float>(m_list, m_array);
    euler.setAngle(m_array);

    delete [] m_array;
    m_array = 0;
    
    return python::detail::none();
}



const char* PyEuler::py_doc(const char* funcname)
{
    if (!funcname) {
	return "";
    }

    static bool has_built = false;
    static python::dictionary dict = python::dictionary();
    
    if (!has_built) {
    
	dict["Euler"] = "Euler();\nEuler(a1, a2, a3); \nEuler(a1, a2, a3, type); \nEuler(e1, e2, e3, e0, type);";
    
	dict["init"] = "init()";    
	dict["rectify"] = "rectify()";
    
	dict["setAngle"] = "setAngle(a1, a2, a3, type = EMAN) \nsetAngle(e1, e2, e3, e0, type) \n setAngle(list matlist): matlist: a matrix with size=9";
    
	dict["setTyp"] = "setTyp(type)";    
	dict["increment"] = "increment(delAlt, delAz, delPhi) \n  increment an Euler in place";    
	dict["diff"] = "float diff(Euler ang, nophi=true) \n  find angle between two direction vectors";
	dict["inverse"] = "Euler inverse() \n  takes the inverse of an euler";    
	dict["fromLeft"] = "fromLeft(Euler leftEuler) \n  rotates a given rotation  on the left \n\n fromLeft(alt, az, phi) \n  rotates a given rotation  on the left";
    
	dict["EulerMult"] = "Euler EulerMult(Euler bc) \n  s the value of the given rotation";    
	dict["getMaxSymEl"] = "int getMaxSymEl() \n  gets the order of the group";    
	dict["setSym"] = "setSym(string symmetry) \n  initialize symcounter to zero";    
	dict["resetSym"] = "resetSym() \n  initializes symcounter_ without ing 1st element";    
	dict["FirstSym"] = "Euler FirstSym() \n  initializes";    
	dict["NextSym"] = "Euler NextSym()";    
	dict["SymN"] = "Euler SymN(n) \n  s the n'th symmetric element";    
	dict["valid"] = "int valid()";    
	dict["fromMRC"] = "fromMRC(theta, phi, omega) \n  set from MRC angles ";    
	dict["thetaMRC"] = "float thetaMRC()";    
	dict["phiMRC"] = "float phiMRC();";    
	dict["omegaMRC"] = "float omegaMRC()";    
	dict["alt"] = "float alt()";
	dict["az"] = "float az()";    
	dict["phi"] = "float phi()";    
	dict["alpha"] = "float alpha()";    
	dict["beta"] = "float beta()";
	dict["gamma"] = "float gamma()";
	dict["phiSpider"] = "float phiSpider()";    
	dict["thetaSpider"] = "float thetaSpider()";    
	dict["gammaSpider"] = "float gammaSpider()";    
	dict["e0"] = "float e0()";
	dict["e1"] = "float e1()";
	dict["e2"] = "float e2()";
	dict["e3"] = "float e3()";
	dict["A"] = "float Q()";
	dict["n1"] = "float n1()";    
	dict["n2"] = "float n2()";    
	dict["n3"] = "float n3()";    
	dict["Qsgi"] = "float Qsgi()";
	dict["n1sgi"] = "float n1sgi()";
	dict["n2sgi"] = "float n2sgi()";
	dict["n3sgi"] = "float n3sgi()";
	dict["mat"] = "list mat()";    
	dict["Print"] = "print(id_str, out=stdout)";    
	dict["printMatrix"] =  "printMatrix(id_str, out=stdout);";    
	dict["doc"] = "doc(funcname) \n  return docstringgiven a function name";

	has_built = true;
    }
    
    const char* fdoc = BOOST_PYTHON_CONVERSION::from_python(dict.get_item(funcname).get(),
							    python::type<const char*>());
    if (!fdoc) {
	fdoc = "Error: not defined";
    }
    return fdoc;
}
    
    
