// batchboxer.C  9/11/01 Steve Ludtke

//N batchboxer
//F This program is used for offline boxing of micrographs (MRC format ONLY!)
//T LM
//
//1
//P [refimg=<filespec>] 
//P [dbbox=<filespec>]	Input box database
//P [dbin=<filespec>]	Read box database from particle image set headers
//P [auto[=<refs>,<thr1>,<thr2>,<thr3>]]	Autobox with defined thresholds
//P [xform=<dx>,<dy>,<da>]	Rotate and translate the box database before output (see alignhuge)
//P [scale=<mult>]	Scale the box locations, useful if the micrograph has been resized
//P [newsize=<size>]	New box size for all boxes (center preserved)
//P [input=<filespec>]	Image to be boxed	(MRC format only!!!)
//P [output=<filespec>]	Output file for boxed images
//P [ctfout=<filespec>]	File to put CTF curves into (if input is focal pair merged)
//P [dbout=<filespec>]	Output box database file
//P [invert]	Invert the particle image before output
//U batchboxer dbbox=jj0441.box input=jj0441.mrc output=0441.hed
//D This program does offline boxing of particles using a box database or an automatic
//E boxing routine. Please note that it will work only on MRC images.

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include <iostream.h>
#include <stdlib.h>

#define CLIP	768

#define DB(x) printf("%s\n",x);

// class for dealing with box databases, the x,y coordinates are 
// stored internally as the center of the box, but returned and
// passed in as the lower left corner of the image
class BOXLIST {
	int *x,*y;
	int n,na;
	int sz;
	
public:
	BOXLIST();
	~BOXLIST();
	void empty();
	void append(int x,int y);
	int atx(int i);
	int aty(int i);
	void setsize(int s);
	int size();
	int count();
	void xform(int dx,int dy,float da,int xsize,int ysize);
	void scale(float sca);
	void write(const char *file);
	void read(const char *file);
	void readimg(const char *file);
};

BOXLIST::BOXLIST() { 
	x=(int *)malloc(8*sizeof(int)); 
	y=(int *)malloc(8*sizeof(int));
	na=8;
	n=0;
}

BOXLIST::~BOXLIST() {
	free(x);
	free(y);
}

int BOXLIST::size() { return sz; }

void BOXLIST::setsize(int s) { sz=s; }

void BOXLIST::empty() { n=0; }

void BOXLIST::append(int X, int Y) { 
	x[n]=X+sz/2; 
	y[n]=Y+sz/2;
	n++; 
	if (n>=na) { 
		na*=2;
		x=(int *)realloc(x,na*sizeof(int));
		y=(int *)realloc(y,na*sizeof(int));
	}
}

int BOXLIST::atx(int i) { if (i<n&&i>=0) return x[i]-sz/2; return 0; }
int BOXLIST::aty(int i) { if (i<n&&i>=0) return y[i]-sz/2; return 0; }

int BOXLIST::count() { return n; }

void BOXLIST::write(const char *file) {
FILE *out = fopen(file,"w");
if (!out) return;
for (int i=0; i<n ; i++) fprintf(out,"%d\t%d\t%d\t%d\n",atx(i),aty(i),sz,sz);
fclose(out);
}

void BOXLIST::read(const char *file) {
FILE *in = fopen(file,"r");
if (!in) return;

char s[128];
int xx,yy;
while (fgets(s,127,in)) {
	if (sscanf(s," %d %d %d",&xx,&yy,&sz)!=3) continue;
	append(xx,yy);
}
fclose(in);

}

void BOXLIST::readimg(const char *file) {
int i,j=0,x0,y0,x1,y1;
char t[80];
EMData r;
int n=fileCount(file);

for (i=0; i<n; i++) {
	r.readImage(file,i,1);
	if (sscanf(r.Name()," %d,%d",&x0,&y0)!=2) {
		if (sscanf(r.Name(),"!$ %d,%s %d,%d",&j,t,&x0,&y0)!=4)  continue;
	}
	x1=r.xSize();
	y1=r.ySize();

	append(x0,y0);
}

sz=x1; 
}

void BOXLIST::xform(int dx,int dy,float da,int xs,int ys) {
float mx0=cos(da);
float mx1=sin(da);

for (int i=0; i<n; i++) {
	float tx= (x[i]-xs/2)*mx0+(y[i]-ys/2)*mx1-dx+(xs/2);
	float ty=-(x[i]-xs/2)*mx1+(y[i]-ys/2)*mx0-dy+(ys/2);
	x[i]=tx;
	y[i]=ty;
}

}

void BOXLIST::scale(float sca) {
for (int i=0; i<n; i++) {
	x[i]*=sca;
	y[i]*=sca;
}
sz*=sca;

}

BOXLIST boxes;
unsigned char *bestmap=NULL;

void autoBox(char *micro,char *ref,float *thresh);
void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize);
EMData *ptclCorr(EMData *data, List &refs,int xo,int y0,int boxsize);

main(int argc,char *argv[]) {
char *input=NULL,*boxed=NULL,*dbout=NULL,*dbin=NULL,*imgin=NULL,*ctfout=NULL,*ref=NULL;
float autothr[4] = { 0,0,0,1.0 };
float xfdx=0,xfdy=0,xfda=0,scale=1.0;
char refs[512];
int i,newsize=0;
int invert=0;

appinit(argc,argv);

if (argc<2) appusage(argv[0]);

for (i=1; i<argc; i++) {	
	if (strncmp(argv[i],"input=",6)==0) input=argv[i]+6;
	else if (strncmp(argv[i],"output=",7)==0) boxed=argv[i]+7; 
	else if (strncmp(argv[i],"ctfout=",7)==0) ctfout=argv[i]+7; 
	else if (strncmp(argv[i],"scale=",6)==0) scale=atof(argv[i]+6); 
	else if (strncmp(argv[i],"newsize=",8)==0) newsize=atoi(argv[i]+8); 
	else if (strncmp(argv[i],"dbout=",6)==0) dbout=argv[i]+6;
	else if (strncmp(argv[i],"dbbox=",6)==0) dbin=argv[i]+6;
	else if (strncmp(argv[i],"imgbox=",7)==0) imgin=argv[i]+7;
	else if (strncmp(argv[i],"auto=",5)==0) sscanf(argv[i]+5,"%f,%f,%f,%f",autothr,autothr+1,autothr+2,autothr+3);
	else if (strncmp(argv[i],"xform=",6)==0) sscanf(argv[i]+6,"%f,%f,%f",&xfdx,&xfdy,&xfda);
	else if (strncmp(argv[i],"invert",6)==0) invert=1;
	else if (strncmp(argv[i],"refimg=",7)==0) ref=argv[i]+7; 
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

 
int Ref=LOGbegin(argc,argv);

boxes.empty();
if (dbin) { printf("Read boxes from %s\n",dbin); boxes.read(dbin); }
if (imgin) { printf("Extract location info from %s\n",imgin); boxes.read(imgin); }
if (autothr[0]||autothr[1]||autothr[2]) {
    //autobox micrographs
    //input: entire micrographs
    //ref:  reference image sets
    //autothr: thresholds
    autoBox(input,ref,autothr);
}

if (boxes.count()==0) error(ERR_EXIT,"No boxes to process");
printf("%d Boxes to process\n",boxes.count());

if (newsize>0) boxes.setsize(newsize);

if (xfdx||xfdy||xfda) {
	if (!input) error(ERR_EXIT,"Cannot perform transformation without an input image.");
	EMData tmp;
	tmp.readImage(input,0,1);
	boxes.xform(xfdx,xfdy,xfda,tmp.xSize(),tmp.ySize());
}
  if (scale!=1.0) boxes.scale(scale);


if (dbout) boxes.write(dbout);
if (boxed) {
	if (!input) error(ERR_EXIT,"Cannot output boxed particles without an input image");
	printf("Boxing %s into %s\n",input,boxed);
	EMData tmp;
	tmp.readMRC(input,1);
	char s[120];
	int ni;
	strcpy(s,tmp.Name());
	for (i=0; i<boxes.count(); i++) {
		tmp.readMRCArea(input,boxes.atx(i),boxes.aty(i),boxes.size(),boxes.size());
		if(invert) tmp.invert();
		char t[120];
		if (s[0]=='!' && s[1]=='$') {
			if (ctfout) {
				char nm[120];
				float apix;
				int u;
				sscanf(s,"!$ %d,%f,%s",&u,&apix,nm);
				EMData tmp2;
				if (i==0) {
					tmp2.readImage(nm,u);
					tmp2.writeIMAGIC(ctfout,-1);
					ni=fileCount(ctfout)-1;
				}
				sprintf(t,"!$ %d,%1.5g,%s %d,%d",ni,apix,ctfout,boxes.atx(i),boxes.aty(i));
			}
			else sprintf(t,"%s %d,%d",s,boxes.atx(i),boxes.aty(i));
		}
		else sprintf(t,"%d,%d",boxes.atx(i),boxes.aty(i));
		tmp.setName(t);
		tmp.writeIMAGIC(boxed,-1);
	}
}

LOG(Ref,NULL,LOG_END,NULL);
}

void autoBox(char *micro,char *ref,float *thresh) {
EMData *d1,*d2,*data;
int nx,ny,i,j;
List *refs;
List abptcl;

// autoboxing option of batchboxer program, very similar to Boxer GUI version

cout<<"thresh[0]="<<thresh[0]<<" thresh[1]="<<thresh[1]<<" thresh[2]="<<thresh[2]<<" thresh[3]="<<thresh[3]<<endl;

//read the micrograph
data=new EMData;
data->readImage(micro,0);
data->normalize();
//data->readMRC(micro,1); 
 
//read ref projection(s) 
refs=readImages(ref,-1,-1); 
d1=(EMData *)refs->objectAt(0);
boxes.setsize(d1->ySize()); 
int boxsize=d1->xSize();
 
if (refs->count()<1) return;
if (refs->count()>250) error(ERR_EXIT,"FATAL ERROR: Max # references 250.");

abptcl.empty();
for (i=0; i<refs->count(); i++) {
	d1=(EMData *)refs->objectAt(i);
	d1->edgeNormalize();
	d2=d1->copy(0,0);
	d2->medianShrink(2);
	d2->unum3=0;
	//	d2->normalize();
	abptcl.addObject(d2);
	delete d1;
}
delete refs;

// moving window
nx=(data->xSize()-boxsize*2)/(CLIP*2-boxsize)+1;
ny=(data->ySize()-boxsize*2)/(CLIP*2-boxsize)+1;

// Starting the main loop for cross correlation and particle extraction
// Both micrograph and refrences will be shrank by a factor of two in ptclCorr
// so take that into account when finding the particle coordinates in boxesFrom

for (i=0; i<ny; i++) {
     for (j=0; j<nx; j++) {

       		d1=ptclCorr(data,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize),boxsize);
		boxesFrom(d1,data,abptcl,j*(CLIP*2-boxsize),i*(CLIP*2-boxsize),0,0,thresh,boxsize);
               
		printf("%d,%d (%d)\n",i,j,boxes.count());

		delete d1;
				    }
				}
}

void boxesFrom(EMData *map,EMData *data,List &abptcl,int x0,int y0,int sx,int sy,float *thresh,int boxsize) {
int i,j,k,l,c,n=0,mn;
float *d,*d2,da1,da2,bd;
EMData *dt,*dtc;

// This function will extract particle locations in the map returned from ptclCorr
// The map contained the peaks value of corr coeff
// keep in mind that the micrograph and refrences have been shrank by a factor of two in ptclCorr

if (!map) return;

//dtc= new EMData;
//dtc->readMRCArea(data->Path(),x0,y0,CLIP*2,CLIP*2);
dtc=data->clip(x0,y0,CLIP*2,CLIP*2);

if (sx<boxsize/4) sx=boxsize/4;
if (sy<boxsize/4) sy=boxsize/4;

mn=4*CLIP*CLIP*4/(boxsize*boxsize);

d=map->getData();
for (j=sy; j<CLIP-boxsize/4; j++) {
	for (i=sx; i<CLIP-boxsize/4; i++) {
		if (d[i+j*CLIP]>thresh[0]) {

		        if (n>mn ||i*2+x0>data->xSize()-boxsize || j*2+y0>data->ySize()-boxsize) continue;
		        dt=dtc->clip(i*2-boxsize/2,j*2-boxsize/2,boxsize,boxsize);

			// check for particles with large 'drawn' areas and eliminate them
			d2=dt->getDataRO();	
			c=0;
			for (l=0; l<boxsize*boxsize; l++) {
				if (d2[l]==0) c++;
			}
			dt->doneData();
			if (c>boxsize*2) {
				delete dt;
				continue;
			}			
		
			// compare candidate particles with reference(s)
			if (thresh[1]<1.0 || thresh[2]>0) {
				EMData *cp=dt->copy(0,0);
				EMData *rf=((EMData *)abptcl.objectAt(bestmap[i+j*CLIP]))->copy(0,0);
				cp->medianShrink(2);

				float mult,add;
				rf->normalizeTo(cp,0,0,&mult,&add);
				cp->subtract(rf);
				
				// now calculate the Kurtosis of the resulting image (0 for gaussians)
				float mean=cp->Mean(),sig=cp->Sigma(),kurt=0;
				float *d=cp->getDataRO();
				for (l=0; l<boxsize*boxsize/4; l++) kurt+=pow((d[l]-mean)/sig,4.0);
				kurt=kurt/(boxsize*boxsize/4.0)-3.0;
				cp->doneData();

				if (fabs(kurt)>thresh[3] || mult<thresh[2] || cp->Sigma()>thresh[1] || mult==1) { delete cp; delete rf; delete dt; continue; }

				delete cp;
				delete rf;
			}

    			// ok, we've got a good particle
			boxes.append(i*2+x0-boxsize/2,j*2+y0-boxsize/2);
			delete dt;
			n++;
		}
	}

}
delete dtc; 
map->doneData(); 
}



EMData *ptclCorr(EMData *data,List &abptcl,int xo,int yo,int boxsize) {
EMData *dc,*dd,*dn,*ccf,*ref,*dt;
int i,j,ii,jj,k,x1,y1,l;
float *dp,*ddp,t,s,sig;

// This function will return best correlation map (peak values) from the micrographs
// using a set of references


float th=0.00001;

//dd=new EMData;
//dd->readMRCArea(data->Path(),xo,yo,CLIP*2,CLIP*2);
dd=data->clip(xo,yo,CLIP*2,CLIP*2);
dd->meanShrink(2);


//sig=dd->Sigma();
//if(sig>th) {dd->addConst(-(dd->Mean()));dd->multConst(1.0/sig);}
//if(sig<=th) {dd->addConst(-(dd->Mean()));}
//if(sig=0){exit(1);}
	
dc=dd->copy(0,0);
dn=dd->copy(0,0);
dd->zero();
dn->filter(-5,CLIP/50.0,1);
dc->subtract(dn);

if (!bestmap) bestmap=(unsigned char *)malloc(CLIP*CLIP);
else bestmap=(unsigned char *)realloc(bestmap,CLIP*CLIP);

for (i=0; i<abptcl.count(); i++) {
        ref=((EMData*)abptcl.objectAt(i))->clip(-(CLIP-boxsize/2)/2,-(CLIP-boxsize/2)/2,CLIP,CLIP);


        //msd set for reference
	//sig=ref->Sigma();
	//if(sig>th) {ref->addConst(-(ref->Mean()));ref->multConst(1.0/sig);}
	//if(sig<=th) {ref->addConst(-(ref->Mean()));}
	//if(sig=0){exit(1);}
	
	
        ref->multConst(1.0/ref->Sigma());
	ref->toCorner();
    
	ccf=dc->calcCCF(ref);

	dp=ccf->getDataRO();
	ddp=dd->getData();
	for (j=0; j<CLIP*CLIP; j++) {
		if (dp[j]>ddp[j]) { ddp[j]=dp[j]; bestmap[j]=i; }
	}
	ccf->doneData();
	dd->doneData();
	
	delete ccf;
	delete ref;
}


// low pass filter so peaks are separated from each other
dd->filter(0,CLIP/6.0,5);
t=dd->Max();
dp=dd->getData();
 
//locate peaks
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
	for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
		k=i+j*CLIP;
		if (dp[k]>dp[k-1] && dp[k]>dp[k+1] && dp[k]>dp[k+CLIP] && dp[k]>dp[k-CLIP] && dp[k]>0 &&
			dp[k]>dp[k-1-CLIP] && dp[k]>dp[k+1-CLIP] && dp[k]>dp[k-1+CLIP] && dp[k]>dp[k+1+CLIP]) {
				dp[k]+=t;
		}
	}
}

 // eliminate boxes which are too close to each other
for(i=boxsize/4; i<CLIP-boxsize/4-1; i++) {
	for (j=boxsize/4; j<CLIP-boxsize/4-1; j++) {
		k=i+j*CLIP;
		if (dp[k]>t) {
			x1=i+boxsize/4-1;
			y1=j+boxsize/4-1;
			for (ii=i-boxsize/4; ii<x1; ii++) {
				for (jj=j-boxsize/4; jj<y1; jj++) {
					if (dp[ii+jj*CLIP]>dp[k]) { dp[k]=0; break; }
				}
			}

		}
	}
}

// eliminate everything but the peaks
for (i=1; i<CLIP-1; i++) {
	for (j=1; j<CLIP-1; j++) {
		k=i+j*CLIP;
		if (dp[k]<=t) dp[k]=0; 
		else {
			dp[k]-=t;
			dp[k]*=100.0;
		}
	}
}
 

dd->doneData();
 
delete dc; 
return dd; 
}

