// classesbymra.C ?/98  Steve Ludtke
// Copyright 1998 Baylor College of Medicine

//N classesbymra
//F This program classifies a set of particles based on a set of references (usually projections).
//T LC
//1
//P <particles>	Raw particle filespec	
//P <refrences>	Reference image filespec
//P [split]	Always use this
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [verbose[=<n>]]	More verbose output
//P [mask=<rad>]	Mask the classified particles, also reduces noise for better classification
//P [refmask]	This will use a mask generated from the reference image. Assumes the 3D model has been sensibly masked.
//P [imask=<rad>]	Inside mask, classified particles are masked for classification, but not in the cls files.
//P [maxshift=<rad>]	Maximum translation during image alignment
//P [precen]	Particles are assumed to be accurately precentered 
//P [sigfilt]	Apply a real space filter so only high amplitude data is used for classification
//P [tree1=<treefile>]	First stage high speed classification, see <a href=projtree.html>projtree</a>
//P [tree2=<treefile>]	Second stage high speed classification
//P [ctfc]	Projections are CTF corrected before comparison
//P [matrix]	Generates similarity matrix for entire data set in matrix.dat
//P [sep=<nspl>]	Puts each raw particle in 'nspl' different classes.
//P [logit=<label>]	This specifies a label to use in the particle.log file. This records particle number classification throughout a refinement.
//P [rfp]	This will classify using 'rotational footprints' instead of the images themselves. Much faster and quite accruate.
//P [norot]	This will not do any rotational alignment in classification. Primarily used with projections and the matrix option, but may have other uses.
//P [usefilt]	Experimental. Use the 'filt' version of the input file for classification.
//P [refine]	Refines 2D alignment done with fast algorithm. Slower, but not 'slow'.
//P [slow]	Uses a very slow, exhaustive search algorithm for alignment
//U  classesbymra start.hed proj.hed split mask=28 ctfc sep=2 tree1=proj.tree
//D This classifies a set of raw particles based on a set of reference projections.
//E The output is a set of cls*.img files, one file for each reference image. If the
//E projections were generated from a CTF corrected model, specify the 'ctfc' option
//E so the projections will be 'decorrected' before classification.
//D The 'tree' options implement 2 part classification for higher speed. Check <a href=projtree.html>projtree</a>
//E details on this.
//D The 'sep' option allows each particle to be placed into several different classes. If the
//E images are very noisy, or if the angular spacing is small, misclassification of some
//E particles will inevitably occur. The sep option makes it very likely that one of the
//E classes each particle ends up in will be the correct one. The classalignall procedure
//E can then be used to obtain sufficient consistency within the set of images actually
//E used in each class average.

//#define DEBUG

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

// a little class for maintaining an ordered list
class Olist {
public:
float *dot;
int *n;
EMData **ali;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	dot=(float *)malloc(N*sizeof(float));
	n=(int *)malloc(N*sizeof(int));
	ali=(EMData **)malloc(N*sizeof(EMData *));
	for (int i=0; i<N; i++) {
		dot[i]=-1.0e20;
		n[i]=-1;
		ali[i]=NULL;
	}
}

~Olist() {
	free(dot);
	free(n);
	for (int i=0; i<N; i++) delete ali[i];
	free(ali);
}

void insert(float vdot, int vn, EMData *vali) {
	int i,j;
	
	for (i=0; i<N; i++) if (dot[i]<vdot) break;

	if (i==N) { if (vali) delete vali; return; }

	if (ali[N-1]) delete ali[N-1];
	for (j=N-1; j>i; j--) {
		dot[j]=dot[j-1];
		n[j]=n[j-1];
		ali[j]=ali[j-1];
	}
	dot[i]=vdot;
	n[i]=vn;
	ali[i]=vali;
}

};

main(int argc, char *argv[]) {
int i,j,k,*rn,x,y,nr,ns,ntk=1,split=0,num=0,denom=1,ver=0,sigfilt=0,rfp=0,norot=0;
int n0,n1,shrink=1,mask=-1,imask=0,precen=0,ctfc=0,maxshift=-1,quiet=0,matrix=0,slow=0;
int refine=0;
List *data,*dataref,*datarefnc,*newref=NULL;
EMData *d,*d2,*d2f,*df=NULL,*dr,*da,*msk,*msk2;
char s[80],t[160],*f2r = argv[1];
float dot,classes=-999.0,norm,norm2;
float nullctf[11] = { 0,0,0,0,0,0,0,0,0,0,0 };
float *ctf;
char *tree1=NULL,*tree2=NULL,*logit=NULL;
FILE *out;
Olist *best;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {        
	if (strncmp(argv[i],"classes=",8)==0) classes=atof(&argv[i][8]);
    else if (strncmp(argv[i],"split",5)==0) split=1;
	else if (strncmp(argv[i],"tree1=",6)==0) tree1=&argv[i][6];
	else if (strncmp(argv[i],"tree2=",6)==0) tree2=&argv[i][6];
	else if (strncmp(argv[i],"logit=",6)==0) logit=&argv[i][6];
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"imask=",6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp("refmask",argv[i],7)==0) mask=-2;
	else if (strncmp("maxshift=",argv[i],9)==0) maxshift=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"num=",4)==0) num=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"denom=",6)==0) denom=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"shrink=",7)==0) shrink=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"refine",6)==0) refine=1;
	else if (strncmp(argv[i],"precen",7)==0) precen=1;
	else if (strncmp(argv[i],"sigfilt",7)==0) sigfilt=1;
	else if (strncmp(argv[i],"norot",5)==0) norot=1;
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else if (strncmp(argv[i],"matrix",6)==0) matrix=1;
	else if (strncmp(argv[i],"ctfc",4)==0) ctfc=1;
	else if (strncmp(argv[i],"slow",4)==0) slow=1;
	else if (strncmp(argv[i],"rfp",3)==0) rfp=1;
	else if (strncmp(argv[i],"sep=",4)==0) ntk=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"usefilt",7)==0) {
		f2r=(char *)malloc(strlen(argv[1])+6);
		strcpy(f2r,argv[1]);
		sprintf(f2r+strlen(argv[1])-4,".filt%s",argv[1]+strlen(argv[1])-4);
	}
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (strncmp(argv[i],"verbose=",8)==0) ver=atoi(&argv[i][8]);
		else ver=1;
	}
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if (tree1 && tree2) { printf("Specify either tree1 or tree2, not both!\n"); exit(1); }
if (num>=denom || num<0) { printf("0 <= num < denom !!!\n"); exit(1); }
if (classes==-999.0 && !split) { printf("Must specify at least 1 of 'classes' or 'split'.\n"); exit(1); }
//if (imask>1 && !precen) { printf("imask may only be used in conjunction with precen\n"); exit(0); }
if (rfp && matrix) error(ERR_WARN,"matrix not properly normalized with rfp");

// READ Reference Images
if (tree1) {
	FILE *in = fopen(tree1,"r");
	List *ndr = new List;
	char s[256];
	int cn;
	dataref=new List;

	if (!in) { printf("Cannot open '%s'\n",tree1); exit(1); }

	while (fgets(s,255,in)) {
		if (sscanf(s," %d",&cn)!=1) continue;

		d=new EMData;
		d->readImage(argv[2],cn);
		d->unum=cn;
		dataref->addObject(d);
		
		if (!num) {
			sprintf(s,"tcls%04d.lst",cn);
			if (access(s,F_OK)) continue;
			unlink(s);
//			sprintf(s,"tcls%04d.img",cn);
//			unlink(s);
		}
	}
	fclose(in);
	nr=dataref->count();
}
else if (tree2) {
	FILE *in = fopen(tree2,"r");
	List *ndr = new List;
	char s[256];
	int cn,cn2,pp[19];
	dataref=new List;

	if (!in) { printf("Cannot open '%s'\n",tree2); exit(1); }

	if (sscanf(argv[1],"tcls%d",&cn2)!=1) {
		printf("tree2 requires 'tcls*' files for input\n"); exit(1);
	}
	while (fgets(s,255,in)) {
		if (sscanf(s," %d",&cn)!=1 || cn!=cn2) continue;

		cn=sscanf(s," %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
			pp,pp+1,pp+2,pp+3,pp+4,pp+5,pp+6,pp+7,pp+8,pp+9,pp+10,pp+11,
			pp+12,pp+13,pp+14,pp+15,pp+16,pp+17,pp+18);

		if (cn<4) { printf("Tree file appears strange, '%s'\n",s); exit(1); }
		
		for (i=0; i<cn; i++) {
			d=new EMData;
			d->readImage(argv[2],pp[i]);
			d->unum=pp[i];
			dataref->addObject(d);
		}
		break;
	}
	fclose(in);
	nr=dataref->count();
}
else {
	nr=fileCount(argv[2]);
	dataref=readImages(argv[2],0,nr);		// read all input images
	if (dataref==NULL) error("Cannot read input",ERR_EXIT);
	nr=dataref->count();
	for (i=0; i<nr; i++) {
		d=(EMData *)dataref->objectAt(i);
		d->unum=i;
	}
}

ns=fileCount(f2r);
printf("%d images to classify.\n",ns);

// Read the file containing the list of images to exclude
FILE *za = fopen("start.exclude","r");
char *zapped=NULL;
if ((strcmp(argv[1],"start.hed")==0 || strcmp(argv[1],"start.img")==0) && za) {
	printf("Using image exclusion file\n");
	zapped=(char *)calloc(1,ns);
	while (fgets(s,79,za)) zapped[atoi(s)]=1;
	fclose(za);
}

if (matrix) {
	if (!num) { 
		int tt[3];
		tt[0]=0;
		tt[1]=nr;
		tt[2]=ns;
		FSsend(FS_BIN,12,"matrix.dat",(char *)tt);
	}
	else sleep(1); // stupid hack to avoid possible file contention
}

printf("Read %d references.\n",nr);

rn=(int *)malloc(nr*sizeof(int));
for (i=0; i<nr; i++) rn[i]=((EMData *)dataref->objectAt(i))->unum;

int Ref=0;
if (!quiet) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[2],LOG_INFILE,NULL);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

for (i=0; i<nr; i++) {
	d=(EMData *)dataref->objectAt(i);
	j=d->unum;
	if (sigfilt) d->realFilter(3,d->Sigma(),-d->Sigma());
	d->edgeNormalize(1);
//	d->normalize();
	d->applyMask(mask,4);
	d->meanShrink(shrink);
	d=d->copy();
	d->unum=j;
	dataref->replaceObjectAt(i,d);
	(d->makeRFP(1))->normalize();
}

x=d->xSize();
y=d->ySize();

// initialize images for classums
if (classes!=-999.0) {
	newref=new List;
	for (i=0; i<nr; i++) {
		d=new EMData;
		d->setSize(x,y,1);
		d->zero();
		newref->addObject(d);
	}
}

msk=((EMData *)dataref->objectAt(0))->copy();
msk->setParent(NULL);
msk->one();
msk->applyMask(mask/shrink,4);
msk->applyMask(imask/shrink,5);

msk2=((EMData *)dataref->objectAt(0))->copy();
msk2->setParent(NULL);
msk2->one();
msk2->applyMask(mask/shrink,4);
msk2->applyMask(imask/shrink,5);


d=new EMData;

norm=x*y/1000.0;
norm2=PI/(4.0*x*y);
if (maxshift<0) {
	if (mask<=0) maxshift=y/8; else maxshift=y/2-mask/shrink;
}

n0=num*ns/denom;
n1=(num+1)*ns/denom;

printf("processing %d - %d of %d\n",n0,n1,ns);
fflush(stdout);
if (mask==-1) mask=-shrink;
if (!quiet) LOG(Ref,"cls0000.lst",LOG_OUTFILE,NULL);

if (ctfc) {
	datarefnc=dataref; 
	dataref=new List; 
	for (i=0; i<nr; i++) {
		dataref->addObject(((EMData *)datarefnc->objectAt(i))->copy());
		((EMData *)dataref->objectAt(i))->setCTF(nullctf);
	}
}

float *sim = (float *)malloc(sizeof(float)*nr);

// align and classify
for (i=n0; i<n1; i++) {
	best=new Olist(ntk);
	if (zapped && zapped[i]) continue;	// This particle has been marked as bad

	d->readImage(f2r,i);
// we set the B-factor to zero so ctf correction doesn't end up applying it twice
// This won't work properly for CTFF images obviously. still needs to be dealt with...
	if (d->hasCTF()) {
		float *par = d->getCTF();
		par[1]=0;
	}
	d->edgeNormalize(1);
//	d->normalize();
	d->meanShrink(shrink);
	d->mult(msk2);
	if (d->xSize()!=x ||d->ySize()!=y) { printf("Bad image size %d,%d != %d,%d\n",d->xSize(),d->ySize(),x,y); exit(1); }

	d2=d->copy();
	d2->makeRFP(1);

//	d2f=d->copy(0,0);
//	d2f->hFlip();
/*	if (ver>=9) {
		d2->writeIMAGIC("align.hed",-1);
		df->writeIMAGIC("align.hed",-1);
	}*/
	
	for (j=0; j<nr; j++) {
		dr=(EMData *)dataref->objectAt(j);
		if ((d->hasCTF()||d->hasCTFF()) && ctfc && d->CTFCmp(dr)) {
			EMData *dt;
			// if several images in a row have the same ctf, don't recalculate
			ctf=d->ctfCurve(2);
//			save_data(0,1,ctf,CTFOS*d->ySize()/2,"c.txt");
			dt=((EMData *)datarefnc->objectAt(j))->doFFT()->copy();
			dt->applyRadFn(CTFOS*d->ySize()/2,0,1.0/CTFOS,ctf,0);
			dr=dt->doIFT();
			dt->gimmeFFT();
			dr->edgeNormalize();
//			dr->normalize();
			dr->mult(msk);
			dr->setParent(NULL);
			dr->setCTF(d->getCTF());
			if (mask==-2) { 
				EMData *cp=((EMData *)datarefnc->objectAt(j))->copy(0,0);
				cp->realFilter(6);	//value!=0 -> 1
				dr->mult(cp);
				delete cp;
			}
			(dr->makeRFP(1))->normalize();
			delete ((EMData *)dataref->replaceObjectAt(j,dr));
			delete dt;
			free(ctf);
		}
		if (rfp) {
			df=d2->copy(0,0);
			df->hFlip();
			df->makeRFP(1);
			dot=d2->rotAlign(dr);
			float dot2=df->rotAlign(dr);
			if (dot2>dot) dot=dot2;
			da=NULL;
			dot/=10;
		}
		else {
			if (precen) {
				Euler e = *dr->getEuler();
				da=dr->RFAlign(d2,NULL,imask/shrink);
				dr->setRAlign(&e);
			}
			else if (norot) { da=dr->copy(); da->transAlign(d2,1,1,maxshift); da->fastTranslate(); }
			else {
				Euler e = *dr->getEuler();
				if (slow) da=dr->RTFAlignSlow(d2,NULL,maxshift);
				else da=dr->RTFAlign(d2,NULL,1,maxshift);
				if (refine) da->refineAlign(d2,1);
				dr->setRAlign(&e);
			}
			if (!da) continue;
			da->mult(msk);
//			dot=da->dot(dr);
#ifndef ALTCMP
			dot=da->lcmp(d2,1);
#else
			dot=(da->ySize()/2-da->fscmp(d2))/(float)da->ySize();
#endif

//if (j==15) dr->makeRFP()->writeIMAGIC("align.hed",-1);

			// This is a new normalization, basically equivalent to lcmp. The problem this corrects is
			// That 'edgenormalize' makes direct comparison of dot products problematical
//			dot=(da->Sigma()+SQR(da->Mean()))*da->xSize()*da->ySize()+(dr->Sigma()+SQR(dr->Mean()))*dr->xSize()*dr->ySize()-2*dot;
//			dot/=da->xSize()*da->ySize();
		}
#ifdef DEBUG
		EMData *DDB,*DDR;
		DDB=d2->copy(0,0);
		DDR=da->copy(0,0);
		DDB->writeIMAGIC("mradbug.hed",-1);
		DDR->writeIMAGIC("mradbug.hed",-1);
		DDR->normalizeTo(DDB,1);
		DDB->subtract(DDR);
		if (mask==-2) { DDR->realFilter(6); DDB->mult(DDR); }
		DDB->writeIMAGIC("mradbug.hed",-1);
//		printf("%d. %f\n",j,DDB->Sigma());
		delete DDB;
		delete DDR;
#endif
//		da->subtract(dr);
//		da->writeIMAGIC("t.hed",-1);
		
		if (da) sim[j]=dot;
		else sim[j]=0;
//		if (sim[j]<0) sim[j]=0;
		if (ver) printf("%d\t%f\t%f\t %1.1f,%1.1f  %7.3f\t%1.5f  %1.5f  %c\n",
			j,dot,sim[j],da->Dx(),da->Dy(),da->alt()*180./PI,da->Sigma(),dr->Sigma(),da->isFlipped()?'F':' ');

		if (ver>=9) { 
			d2->makeRFP()->writeIMAGIC("align.hed",-1);
			df->makeRFP()->writeIMAGIC("align.hed",-1);
//			d2->writeIMAGIC("align.hed",-1);
//			df->writeIMAGIC("align.hed",-1);
//			dr->writeIMAGIC("align.hed",-1); 
//			da->writeIMAGIC("align.hed",-1); 			
		}

		if (!rfp) dot=(2.0-dot)*500.0;	// this is for historical purposes
		best->insert(dot,j,da);	// note that this may delete da
		da=NULL;
		
//		if (ver) { putchar('0'+(int)floor(dot/norm/50)); fflush(stdout); }
	}
	
	if (matrix) {
		int *tt = (int *)malloc((2+nr)*sizeof(int));
		tt[0]=-1;
		tt[1]=i;
		memcpy(tt+2,sim,nr*sizeof(int));
		FSsend(FS_BIN,(2+nr)*sizeof(int),"matrix.dat",(char *)tt);
		free(tt);
	}
	
//	if (ver) putchar('\n');

/*	if (best->dot[0]<0) { 
		printf("No class for %d\n",i);  
		delete d2;
		delete df;
//		delete d2f;
		continue;
	}*/

	delete d2;
//	delete d2f;
	if (df) { delete df; df=NULL; }

	// class average for dots better than a threshold	
//	if (classes!=-999.0) {
//		if (bestdot/norm>classes) {
//			d2=(EMData *)newref->objectAt(best);
//			d2->add(dbest);
//		}
//	}

	// split into class files
	if (split) {
/*		if (shrink>1) d->readImage(argv[1],i);
		printf("%d -> ",i);
		for (j=0; j<ntk; j++) {
			printf("\t%d (%1.2f)",rn[best->n[j]],best->dot[j]);
			sprintf(s,"%scls%04d.hed",tree1?"t":"",rn[best->n[j]]);
			d->writeIMAGIC(s,-1);
		}
		printf("\n");
		fflush(stdout);*/

		printf("%d -> ",i);
		for (j=0; j<ntk; j++) {
			printf("\t%d (%1.2f)",rn[best->n[j]],best->dot[j]);
			sprintf(s,"%scls%04d.lst",tree1?"t":"",rn[best->n[j]]);
			sprintf(t,"%1.2f",best->dot[j]);
			d->writeLST(s,-1,argv[1],i,t);
			
			if (logit) {
				// order: program tag, particle number, ntk rank, quality, class, iteration tag 
				char s[120];
				sprintf(s,"C\t%d\t%d\t%1.5g\t%d\t%s\n",i,j,best->dot[j],rn[best->n[j]],logit);
				FSsend(FS_TEXT,strlen(s)+1,"particle.log",s);
			}
		}
		printf("\n");
		fflush(stdout);
		
	}
	
	delete best;
}



if (classes!=-999.0) {
	for (i=0; i<nr; i++) {
		d2=(EMData *)newref->objectAt(i);
//		d2->normalize();
		d2->writeIMAGIC("classes.hed",-1);
	}
	printf("Class averages written\n");
}

if (!quiet) LOG(Ref,NULL,LOG_END,NULL);
		
}



