// helixhunter2.C by Wen Jiang, May 2000
// Copyright 2000 Baylor College of Medicine
// 
//
//N helixhunter2
//F This program is used to find possible alpha-helices region in medium resolution (8Angstrom) density map
//T X
//1
//P <input>	input file, single 3D map in any format that EMAN supports, usually the reconstruction map
//P	<output>	output file, Open Inventor format, explicitly describing each helix 
//P	<a/pix>	the sampling for input file, Angstrom/pixel 
//P	[minlen=<Angstrom>]	the minimal length of density region to qualify for a helix, in Angstrom
//P	[maxlen=<Angstrom>]	the maximal length of density region to qualify for a helix, in Angstrom
//P	[hlxnum=<int>]	the number of helices wanted, the correct threshhold would be automatically searched
//P	[sigma=<number>]	threshhold set to number of sigma above mean for pre-threshholding the map
//P	[percen=<0-1>]	threshhold set to percent of the maximal density for pre-threshholding the map, this is prefered when both are set
//P	[view]	will automatically call ivview to show the result Open Inventor file 
//P	[list=<filename>]	also output a simple text file describing the helices: 9 white space delimited parameters for each helix(order: 3 center, 4 quaternions, radius, length)
//P	[dejavu=<filename>]	also output a DejaVu secondary structure elements file, see DejaVu document for more details 
//P	[cosec=<filename>]	also output a COSEC secondary structure element file, see COSEC document for more details
//P	[explorer=<binary file>]	output a file to read by an IRIS Explorer module to apply the helices
//P	[sortbyvalue]	sort by the pixel value at each helix center instead of the default length criterion
//P	[relorig=<x,y,z>]	the position of the small boxed density(model) in the large map, default to 0,0,0
//P	[corner]	leave the positions in the first octant, default is to center helices 
//P	[verbose]	show intermediate output for monitoring progress and more working details 
//P	[docylccffirst]	to do exhaustive cross correlation search with prototypic helix first, following parameters are useful only if this is set
//P	[cyl]	output a protoypic helix density map in MRC format 
//P	[int=<file>]	prefix for the 3 output MRC maps from the resulted cross correlation search, 
//P	[da=<angle>]	angular step size for the search, default to appropriate number for the resolution and map size
//P	[res=<Angstrom>]	resolution of the density map, default to 8 Angstrom
//P	[type=<0-2>]	type of prototypic helix, default to 2
//P	[len=<number>]	the length of prototypic helix, default to 16.2 Angstrom (3turns)
//P	[ccf]	using straight cross correlation function instead of the default MCF(mutual correlation function)
//P	[vectavg[=<angstrom>]]	the score map will be locally vector averaged after the seach, default to 3x3x3 Angstrom kernel size
//U	helixhunter2 input.mrc output.iv 1.5 
//U	helixhunter2 input.mrc output.iv 1.5 verbose view cosec=out.cosec dejavu=out.dejavu list=helices.txt percen=0.8 
//U	helixhunter2 input.mrc output.iv 1.5 verbose view docylccffirst da=5 percen=0.5
//D This program is still somewhat experimental.
//D	This programs tries to find the helices density regions in a medium resolution density map. 
//E There are two modes for this program: the default is to work directly on the density map 
//E by segmenting the high density regions, analyzing the properties for each region, and making 
//E decisions if the region is qualitied as a helix; the second mode is to pre-process the density map by
//E	exhaustive cross correlating with prototypic helix at all possible orientations, re-weighting 
//E the original densities, and then adopt the first mode to find the helices. While the default 
//E mode is very fast which usually takes less than 10 seconds and the second mode would take 
//E much longer time which usually in minutes to tens of minutes range, the second mode usually 
//E gives superior results. The parameters that need to pay attention to are the two threshhold 
//E parameters <b>percen</b> and <b>sigma</b>, the higher the threshhold, the less numer of and 
//E the longer helices would be found, and the lower the threshhold, the more numer of and the shorter 
//E	and ofter non-helix regions are found.
//D	Visit web page <a href="http://ncmi.bcm.tmc.edu/~wjiang/hhfh">Helixhunter and Foldhunter</a> for 
//E	more information.

#include <string>
#include <iostream.h>
#include "Cylinder.h"

void	LocalVectorAverage(EMData*best, EMData*alt, EMData*az, float mask_width);
Cylinders * DensityToCylinder(	EMData *in, float thresh, bool verbose = false, float maxrad = 2.5, 
								float minlen = 8.1, float maxlen = 100000., float min_len_width_ratio = 2, 
								float max_width_width_ratio = 1.75);

int	main(int argc, char *argv[])
{

	appinit(argc, argv);

	EMData	 * model, *best, *balt, *baz;
	int	nz, i, j, k, l, ps, ps2, type = 2;
	int	ii, jj, kk, nn = 0, kcyl = 0, hlxnum=-1;
	float	alt, az, dlaz, apix, x, y, z, s, bm, bs, sigallow = 0.5;
	float	*dt, *dt2, *dt3;
	float	*a[4], dl;
	char	*intfile = NULL;
	float	DA = -1;	//(5.0*PI/180.0);
	float	res = -1;	// default to 8angstrom, won't be effective if da is set
	float	vectwidth = -1, length = -1;
	bool	docylccffirst = false, vectavg = false, useccf = false;
	char	*listfile = NULL, *dejavufile = NULL, *cosecfile = NULL;
	char	*expfile = NULL;	// if output binary file for an explorer module
	float	minlen = -1, maxlen=-1, percen = -1, sigma = -1;
	float rox=0, roy=0, roz=0;
	bool	view = false, sortweight = false, verbose = false, corner = false, relorig=false;

	if (argc < 4) appusage(argv[0]);

	for (i = 4; i < argc; i++) {
		if (strncmp(argv[i], "minlen=", 7) == 0)
			minlen = atof(&argv[i][7]);
		else if (strncmp(argv[i], "maxlen=", 7) == 0)
			maxlen = atof(&argv[i][7]);
		else if (strncmp(argv[i], "hlxnum=", 7) == 0)
			hlxnum = atoi(&argv[i][7]);
		else if (strncmp(argv[i], "sigma=", 6) == 0)
			sigma = atof(&argv[i][6]);
		else if (strncmp(argv[i], "percen=", 7) == 0)
			percen = atof(&argv[i][7]);
		else if (strncmp(argv[i], "list=", 5) == 0)
			listfile = &argv[i][5];
		else if (strncmp(argv[i], "dejavu=", 7) == 0)
			dejavufile = &argv[i][7];
		else if (strncmp(argv[i], "cosec=", 6) == 0)
			cosecfile = &argv[i][6];
		else if (strncmp(argv[i], "explorer=", 9) == 0)
			expfile = &argv[i][9];
		else if (strncmp(argv[i], "sort", 4) == 0)
			sortweight = true;
		else if (strncmp(argv[i], "relorig=", 8) == 0) 
			sscanf(&argv[i][8],"%f,%f,%f",&rox,&roy,&roz);
		else if (strncmp(argv[i], "corner", 6) == 0)
			corner = true;
		else if (strncmp(argv[i], "view", 4) == 0)
			view = true;
		else if (strncmp(argv[i], "verbose", 7) == 0)
			verbose = true;
		else if (strncmp(argv[i], "docylccffirst", 13) == 0)
			docylccffirst = true;
		else if (strncmp(argv[i], "int=", 4) == 0)
			intfile = &argv[i][4];
		else if (strncmp(argv[i], "da=", 3) == 0)
			DA = atof(&argv[i][3]) * PI / 180.0;
		else if (strncmp(argv[i], "cyl", 3) == 0)
			kcyl = 1;
		else if (strncmp(argv[i], "vectavg=", 8) == 0) {
			vectavg = true;
			vectwidth = atof(&argv[i][8]);
		} 
		else if (strncmp(argv[i], "vectavg", 7) == 0)
			vectavg = true;
		else if (strncmp(argv[i], "type=", 5) == 0) {
			type = atoi(&argv[i][5]);
			if (type < 0 || type > 2) {	// make sure in [0,2]
				type = 2;
				fprintf(stderr, "%s:\tTYPE must be in [0,2] in \"%s\", and is set to 2\n", argv[0], argv[i]);
			}
		} 
		else if (strncmp(argv[i], "len=", 4) == 0)
			length = atof(&argv[i][4]);
		else if (strncmp(argv[i], "res=", 4) == 0)
			res = atof(&argv[i][4]);
		else if (strncmp(argv[i], "ccf", 3) == 0)
			useccf = true;
		else {
			cerr << argv[0] << ":\tunknown parameter '" << argv[i] << "'\n";
			exit(1);
		}
	}

	apix = atof(argv[3]);
	if (apix <= 0) {
		fprintf(stderr, "A/pix must be >=0\n");
		exit(1);
	}
	if (apix > 2)  {
		fprintf(stderr, "Helixhunter will not work properly if A/pix > 2\n");
	}
	if (apix > 1.5) {
		fprintf(stderr, "Helixhunter works better if A/pix < 1.5\n");
	}

	if (percen <= 0 && sigma) {
		if(docylccffirst) 	{
			percen = 0.45;
		}
		else {
			percen = 0.85;
		}
		cerr << argv[0] << ":\tNone of [sigma=<number>] or [percent=<0-1>] is set, so \"percen\" is set to " << percen << "\n";
	} else if (percen > 0 && sigma > 0) {
		cerr << argv[0] << ":\tOnly ONE of [sigma=<number>] and [percent=<0-1>] should be set, so \"percen=" << percen
		     << "\" is used, and \"sigma=" << sigma << "\" is ignored\n";
		sigma = -1;
	} // now only one of percen or sigma is valid	

	if(minlen<=0) {
		cerr << argv[0]<<":\t[minlen=<angstrom>] is not set, default to 8.1 angstrom\n";
		minlen=8.1;
	}
	
	if(maxlen<=0) {
		maxlen=100000;	// arbitrary large number
	}
	
	model = new EMData;
	if (model->readImage(argv[1], -1)) {
		cerr << argv[0] << ":\terror reading image " << argv[1] << endl;
		exit(1);
	}
	model->setPixel(apix);
	model->multConst(1.0 / model->Max());

	int	Ref = LOGbegin(argc, argv);
	LOG(Ref, argv[1], LOG_INFILE, NULL);

	if (docylccffirst) {
		//model->normalize();
		threshHold(model, model->Mean() + 2 * model->Sigma());
		model->setPixel(apix);

		if (vectwidth <= 0)
			vectwidth = 3;
		if (length <= 0) {
			cerr << "The protoypic helix length is not set, default to 16.2 Angstrom\n";
			length = 16.2;	//two turns
		}

		nz = model->zSize();
		if (nz != model->xSize() || nz != model->ySize()) {
			cerr<< "The model MUST be in a cubic box (nx=ny=nz)\n";
			exit(1);
		}

		if (DA <= 0) {
			if (res <= 0)
				res = 8;
			DA = 2 * asin(res / nz);
			cerr << "da is set to " << DA * 180.0 / PI << " degree\n";
		}

		best = new EMData;
		best->setSize(nz, nz, nz);
		best->setPixel(apix);
		best->zero();

		balt = new EMData;
		balt->setSize(nz, nz, nz);
		balt->zero();

		baz = new EMData;
		baz->setSize(nz, nz, nz);
		baz->zero();

		EMData * cyl = new EMData;
		cyl->setSize(nz, nz, nz);
		cyl->setPixel(apix);

		for (alt = 0; alt < PI / 2. + DA / 2.0; alt += DA) {
			float	tmp = (sin(alt)*PI / DA);
			dlaz = PI / (float)(tmp);
			for (az = 0; az < 2 * PI + dlaz / 2; az += dlaz) {
				// generate cylinder for ccf
				cyl = modelHelix(cyl, alt, az, type, length);
				if (kcyl && az == 0 && alt == 0) {
					cyl->writeMRC("cyl.mrc");
				}

				EMData * ccf;
				if (useccf) {
					// now do a CCF
					ccf = model->calcCCF(cyl);
				} else {
					ccf = model->calcMCF(cyl);
				}
				ccf->toCorner();

				// Now keep track of the best cylinder at each voxel
				float	*d = ccf->getDataRO();
				float	*dbest = best->getData();
				float	*dalt = balt->getData();
				float	*daz = baz->getData();

				for (k = 0; k < nz; k++) {
					for (j = 0; j < nz; j++) {
						for (i = 0; i < nz; i++) {
							if (*d > 	*dbest) {
								*dbest = *d;
								*dalt = alt;
								*daz = az;
							}

							d++;
							dbest++;
							dalt++;
							daz++;
						}
					}
				}
				ccf->doneData();
				best->doneData();
				balt->doneData();
				baz->doneData();

				if(verbose) {
					printf("alt= %.2f\taz= %.2f\n", alt * 180.0 / PI, az * 180.0 / PI);
					fflush(stdout);
				}
				delete ccf;
			}
		}

		if(verbose)  printf("ccf max=%g min=%g mean=%g sig=%g\n", best->Max(), best->Min(), best->Mean(), best->Sigma());

		if (vectavg)
			LocalVectorAverage(best, balt, baz, vectwidth);	// would modify best contents
		best->multConst(1.0 / best->Max());
		//best->writeMRC("ccf.mrc");

		//best->multConst(1.0/best->Sigma());
		//if (res>=1.0) model->filter(-5,2*apix*model->ySize()/res,2);
		//best->add(model);
		//model->realFilter(2,best->Mean()+best->Sigma()*sigallow);
		//model->realFilter(2,model->Mean()+model->Sigma()*sigallow);
		model->multConst(1.0 / model->Max());
		best->mult(model);
		best->multConst(1.0 / best->Max());
		if(verbose)  printf("ccf max=%g min=%g mean=%g sig=%g\n", best->Max(), best->Min(), best->Mean(), best->Sigma());

		if (intfile) {
			int	size = strlen(intfile);
			best->writeMRC(strcat(intfile, "-coeff.mrc"));
			intfile[size] = 0;
			balt->writeMRC(strcat(intfile, "-alt.mrc"));
			intfile[size] = 0;
			baz->writeMRC(strcat(intfile, "-az.mrc"));
			LOG(Ref, intfile, LOG_OUTFILE, "intermediate");
		}

		delete model;
		model = best;
	}

	Cylinders *cyls;
	if (percen > 0)
		cyls = DensityToCylinder(model, model->Max() * percen, verbose, minlen, maxlen);
	else
		cyls = DensityToCylinder(model, model->Mean() + model->Sigma() * sigma, verbose, minlen, maxlen);
	
	if(hlxnum>0 && cyls->size()!=hlxnum){
		if (percen > 0) {
			float step=percen/5;
			for(int j=0;j<4;j++){
				while (cyls->size()>hlxnum) {
					percen+=step;
					cyls = DensityToCylinder(model, model->Max() * percen, verbose, minlen, maxlen);
				}
				step/=4;
				if( cyls->size()==hlxnum ) break;
				
				while (cyls->size()<hlxnum) {
					percen-=step;
					cyls = DensityToCylinder(model, model->Max() * percen, verbose, minlen, maxlen);
				}
				step/=4;
				if( cyls->size()==hlxnum ) break;
			}
			
		} else {
			float step=sigma/5;
			for(int j=0;j<4;j++){
				while (cyls->size()>hlxnum) {
					sigma+=step;
					cyls = DensityToCylinder(model, model->Mean() + model->Sigma() * sigma, verbose, minlen, maxlen);
				}
				step/=4;
				if( cyls->size()==hlxnum ) break;
				
				while (cyls->size()<hlxnum) {
					sigma-=step;
					cyls = DensityToCylinder(model, model->Mean() + model->Sigma() * sigma, verbose, minlen, maxlen);
				}
				step/=4;
				if( cyls->size()==hlxnum ) break;
			}
		}				
	}		

	if(corner) {
		float tx=(model->xSize()/2-0.5)*apix;
		float ty=(model->ySize()/2-0.5)*apix;
		float tz=(model->zSize()/2-0.5)*apix;
		Vect trans(tx,ty,tz), zero(0,0,0);
		Euler ident(0,0,0);
		cyls->Transform(ident,trans,zero);
	}
	
	if(sortweight) cyls->SortWeight();
	else cyls->SortLength();

	if (verbose && hlxnum>0 ){
		if(percen>0) {
			printf("Final \"percen\" is %g to get %d helices\n",percen,hlxnum);
		} else {
			printf("Final \"sigma\" is %g to get %d helices\n",sigma,hlxnum);
		}
	}
	
	if(verbose)	cyls->WriteListfile(stdout);
	
	if (listfile) {
		cyls->WriteListfile(listfile);
	}
	
	if (dejavufile) {
		cyls->WriteDejaVu(dejavufile);
	}

	if (cosecfile) {
		cyls->WriteCosec(cosecfile);
	}
	if(expfile){
		Helix *helices=new Helix[cyls->size()];
		memset(helices,0,sizeof(Helix)*cyls->size());
		
		for(i=0; i<cyls->size(); i++){
			Euler tmp=Euler(PI/2., 0, 0)*(*cyls)[i]->Axis().euler();		
			Quater quater(tmp);		
			helices[i].qx=quater.nx;
			helices[i].qy=quater.ny;
			helices[i].qz=quater.nz;
			helices[i].qa=quater.Q;
			helices[i].tx=(*cyls)[i]->Xc();
			helices[i].ty=(*cyls)[i]->Xc();
			helices[i].tz=(*cyls)[i]->Xc();
			helices[i].len=(*cyls)[i]->Length();
			helices[i].c =(*cyls)[i]->Weight();
		}
			
		HHH	hhheader;
		memset(&hhheader,0,sizeof(HHH));
		
		strcpy(hhheader.version,"#helixhunter V1 binary\n");        
		hhheader.n=cyls->size();          //number of foldhunter matches 

    	if(strstr(argv[1],"/"))	strcpy(hhheader.fileRef,argv[1]); 
		else	sprintf(hhheader.fileRef,"%s/%s",getenv("PWD"),argv[1]);
		
    	if(strstr(argv[2],"/"))	strcpy(hhheader.fileProbe,"");	// emtpy for helixhunter 
		
    	if(!intfile)	strcpy(hhheader.fileCoeff,"");
		else if(strstr(intfile,"/"))	sprintf(hhheader.fileCoeff,"%s-coeff.mrc",intfile);
		else	sprintf(hhheader.fileCoeff,"%s/%s-coeff.mrc",getenv("PWD"),intfile);
		
		hhheader.apix=apix;
		
		if(relorig) {
			hhheader.rox=rox;	// should be in Angstrom
			hhheader.roy=roy;
			hhheader.roz=roz;
		}
		else {
			hhheader.rox=0;
			hhheader.roy=0;
			hhheader.roz=0;
		}
		
		if(corner){
			hhheader.pcx=0;
			hhheader.pcy=0;
			hhheader.pcz=0;
		}
		else {
			hhheader.pcx=(model->xSize()/2-0.5)*apix;
			hhheader.pcy=(model->ySize()/2-0.5)*apix;
			hhheader.pcz=(model->zSize()/2-0.5)*apix;
		}
			
		FILE* expfp = fopen(expfile,"wb");
		fwrite(&hhheader,sizeof(HHH),1,expfp);	//write the fixed header
		fwrite(helices,sizeof(Helix),cyls->size(),expfp);	// write the variable number of helices
		fclose(expfp);
	} 


	if (!strcmp(&argv[2][strlen(argv[2])-4],".wrl")){
		cyls->WriteVRML(argv[2]);
	} 
	else {
		cyls->WriteInventor(argv[2]);
	}
		
	if (view) {
		string cmd = string("ivview ") + string(argv[2]) + string(" -title ") + string(argv[2]) + string(" &");
		run((char *)(cmd.c_str()));
	}

	LOG(Ref, argv[2], LOG_OUTFILE, NULL);
	LOG(Ref, NULL, LOG_END, NULL);
}


Cylinders * DensityToCylinder(EMData *in, float thresh, bool verbose, float minlen, float maxlen, float maxrad, float min_len_width_ratio, float max_width_width_ratio)
{
	vector < Point * > goodPoints;
	int	nx = in->xSize(), ny = in->ySize(), nz = in->zSize();
	EMData * cp = new EMData;
	cp->setSize(nx, ny, nz);
	cp->one();
	cp->multConst(-1);

	float apix=in->Pixel();
	float	*inp = in->getDataRO(), *cpp = cp->getData();
	int	i, j, k;
	//threshhold the map, only keep significant voxels for further processing
	for (k = 0; k < nz; k++) {
		for (j = 0; j < ny; j++) {
			for (i = 0; i < nx; i++) {
				if (inp[k*nx*ny+j*nx+i] > thresh) {
					goodPoints.push_back(new Point(i, j, k));
				}
			}
		}
	}

	if (verbose)
		printf("%d voxels left and ", goodPoints.size()	);
	vector < vector < Point * > > allobjects;
	vector < Point * > object;

	//first round assignment, group connected voxels into different objects
	for (i = 0; i < goodPoints.size(); i++) {
		bool taken = false;
		for (j = 0; !taken && j < allobjects.size(); j++) {
			for (k = 0; !taken && k < allobjects[j].size(); k++) {
				if (goodPoints[i]->isNeighbor(allobjects[j][k])) {
					allobjects[j].push_back(goodPoints[i]);
					int	x = (int)goodPoints[i]->X(), y = (int)goodPoints[i]->Y(), 	z = (int)goodPoints[i]->Z();
					cpp[z*nx*ny+y*nx+x] = j;
					taken = true;
				}
			}
		}
		if (!taken) { //no suitable objects yet
			vector < Point * > tmp;
			tmp.push_back(goodPoints[i]);
			allobjects.push_back(tmp);
			int	x = (int)goodPoints[i]->X(), y = (int)goodPoints[i]->Y(), 	z = (int)goodPoints[i]->Z();
			cpp[z*nx*ny+y*nx+x] = allobjects.size() - 1;
		}
	}

	//statistics of each object
	if (verbose)
		printf("grouped into %d objects\n", allobjects.size());
	vector < ObjectStat * > objectstat(allobjects.size());
	for (i = 0; i < allobjects.size(); i++) {
		objectstat[i] = new ObjectStat;
		objectstat[i]->SetObject(in, allobjects[i]);
		if (verbose)
			objectstat[i]->print();
	}

	Cylinders *cyls=new Cylinders;
	for (i = 0; i < objectstat.size(); i++) {
		if ((objectstat[i]->len >= minlen) && (objectstat[i]->len <= maxlen) && 
			(objectstat[i]->rad <= maxrad) &&  
		    (objectstat[i]->aspectratio[0] >= min_len_width_ratio) && 
		    (objectstat[i]->aspectratio[1] >= min_len_width_ratio) && 
		    (objectstat[i]->aspectratio[2] <= max_width_width_ratio ) ) {
			Cylinder * cyl = new Cylinder(Point(objectstat[i]->xc, objectstat[i]->yc, objectstat[i]->zc),
			    objectstat[i]->axis, objectstat[i]->len);
			
			int ix=int(objectstat[i]->xc/apix-0.5+nx/2);
			int iy=int(objectstat[i]->yc/apix-0.5+ny/2);
			int iz=int(objectstat[i]->zc/apix-0.5+nz/2);
			float w=in->valueAt(ix,iy,iz);
			
			//printf("weight %d=\t%f\n",cyls->size(),w);
			cyl->setWeight(w);
			
			cyls->add(cyl);
		}
	}
	in->doneData();
	cp->doneData();

	if (verbose)
		printf("But only %d qualified as helices\n", cyls->size())																														;
	return  cyls;
}


void	LocalVectorAverage(EMData*best, EMData*alt, EMData*az, float mask_width)
{
	int	hlfw = round(mask_width / best->Pixel() / 2.);	// how many pixels to average
	if (hlfw == 0)
		return;

	int	nx = best->xSize(), 	ny = best->ySize(), 	nz = best->zSize();
	EMData * best2 = best->copy();
	best->zero();
	float	*bestp = best->getData();
	float	*best2p = best2->getDataRO(), 	*altp = alt->getDataRO(), 	*azp = az->getDataRO();
	for (int k = 0; k < nz; k++) {
		for (int j = 0; j < ny; j++) {
			for (int i = 0; i < nx; i++) {
				int	pos = k*nx*ny + j*nx + i;
				float	vx, vy, vz;
				/*
				vx=sin(altp[pos])*cos(azp[pos])*best2p[pos];
				vy=sin(altp[pos])*sin(azp[pos])*best2p[pos];
				vz=cos(altp[pos])*best2p[pos];
				float coeff;	
				coeff=0.0;
				*/
				vx = vy = vz = 0;
				for (int kk = k - hlfw; kk <= k + hlfw; kk++) {
					for (int jj = j - hlfw; jj <= j + hlfw; jj++) {
						for (int ii = i - hlfw; ii <= i + hlfw; ii++) {
							int	iii = (ii + nx) % nx, 	jjj = (jj + ny) % ny, 	kkk = (kk + nz)
							 % nz;
							int	pos2 = kkk*nx*ny + jjj*nx + iii;
							float	vx2 = sin(altp[pos2])*cos(azp[pos2])*best2p[pos2];
							float	vy2 = sin(altp[pos2])*sin(azp[pos2])*best2p[pos2];
							float	vz2 = cos(altp[pos2])*best2p[pos2];
							/*
							float ang=acos(vx*vx2+vy*vy2+vz*vz2);
							if(ang>PI/2.0)ang=PI-ang;	// make sure 0<angle<PI/2.
							coeff+=best2p[pos2]*cos(2*ang);	// cos weighting, penalizing the difference in orientation 
							*/
							vx += vx2;
							vy += vy2;
							vz += vz2;
						}
					}
				}
				//bestp[pos]=coeff;
				bestp[pos] = sqrt(vx * vx + vy * vy + vz * vz);
			}
		}
	}
	best->doneData();
	alt->doneData();
	az->doneData();
	best2->doneData();
	delete best2;
	//printf("Local vector averaged\n");
}


